#include "main.h"
#include "font.h"
#include "digits.h"

/* Private typedefs */
/* Private defines */
#define FRAMEBUFFER_ROTATE

/* Private constants */
#define MAX_COLUMN    (DISPLAY_COLUMNS - 1)
#define FONT_WIDTH    6
#define STR_BUF_SYM   50
#define STR_BUF_PAD   (DISPLAY_COLUMNS / FONT_WIDTH)
#define STR_BUF_SIZE  ((STR_BUF_PAD + STR_BUF_SYM + STR_BUF_PAD) * FONT_WIDTH)

/* private macros */
/* Private variables */
uint8_t display_Buffer[DISPLAY_COLUMNS] = {0};
static struct Running_Sring {
  uint8_t Buf[STR_BUF_SIZE];
  int Repeat;
  int k;
  int j;
} String = {0};

uint32_t str_Repeat = 0;

/* Privae fuctions */
static void _display_WriteBits(uint16_t data, uint16_t nbits);
static void _delay_c(uint32_t cycle);

/* Exported fuctions */

/**
 * @brief Initialization HT1632C
 */
void display_Init(void) {
  /* stop display if running */
  TIM16->CR1 &= ~(TIM_CR1_CEN);
  Flag.DISP_BSY = 0;

  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();
  HT1632C_CS_ON;

  // Turn on system oscillator
  _display_WriteBits(0x802, 0x800);
  // Turn on LED duty cycle generator
  _display_WriteBits(0x6, 0x100);

  HT1632C_CS_OFF;
  GPIO_SPI_HW();

  display_Fill(0x0);
}

/**
 * @brief Out to display one symbol from specified column
 * 
 * @param symb Symbol to display
 * @param column column from 0 to DISPLAY_COLUMNS-1
 */
void display_Char(const uint8_t symb, const uint8_t column) {
  int i=0, c = column;

  if (c > MAX_COLUMN) {
    c = MAX_COLUMN;
  }

  while ((i < FONT_WIDTH) && (c < DISPLAY_COLUMNS)) {
    display_Buffer[c] = Font_6x8[symb][i];
    i ++;
    c ++;
  }
}

/**
 * @brief Out string to display. Running string. Left shit.
 * @param string String to display, null terminated.
 * @param repeat Times to repeat.
 */
void display_String(const char * string, uint8_t repeat) {
  if (repeat == 0) { return; }
  if (string[0] == 0) { return; }

  display_Wait4Display();
 
  String.Repeat = repeat; // charge global counter
  String.j = DISPLAY_COLUMNS;

  int i = 0, k=0;
  while ((string[i] != '\0') && (i < STR_BUF_SYM)) {
    for(k=0; k<FONT_WIDTH; k++) {
      String.Buf[String.j] = Font_6x8[(uint8_t)string[i]][k];
      String.j ++;
    }
    i ++;
  }
  for (i=0; i<DISPLAY_COLUMNS; i++) {
    String.Buf[String.j + i] = 0x0;
  }

  String.k = 0;
  String.j ++;
  Flag.DISP_BSY = 1;
  TIM16->EGR = TIM_EGR_UG; // force timer update
  TIM16->CR1 |= TIM_CR1_CEN; // start tim16 counter
  TIM16_IRQHandler();
}

/**
 * @brief Set HT1632C PWM Value
 * @param pwm value in 0-15
 */
void display_PWM(uint8_t pwm) {
  // check value
  if (pwm > 15) {
    pwm = 15;
  }
  // align value
  pwm <<= 1;

  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits((0x940|pwm), 0x800);
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Set HT1632C Blink state
 * @param state 0 for off, any other value for on
 */
void display_BlinkState(dis_en_t state) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  if (state == Disable) {
    _display_WriteBits(0x810, 0x800); // Blink off
  } else {
    _display_WriteBits(0x812, 0x800); // Blink on
  }
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Set HT1632C Led state
 * @param state 0 for off, any other value for on
 */
void display_LedState(dis_en_t state) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  if (state == Disable) {
    _display_WriteBits(0x804, 0x800); // Led off
  } else {
    _display_WriteBits(0x806, 0x800); // Led on
  }
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Primitive delay
 */
static void _delay_c(uint32_t cycle) {
  while (cycle) {
    __NOP();
    cycle --;
  }
}

/**
 * @brief Write <nbits> bit of data to selected HT1632Cs
 * @param data words to write
 * @param nbits num of bits (1<<(num-1))
 */
void _display_WriteBits(const uint16_t data, uint16_t nbits) {
//  _delay_c(2);
  do {
    HT1632C_WR_LOW;

    if (data & nbits) {
      HT1632C_DATA_1;
    } else {
      HT1632C_DATA_0;
    }
//    _delay_c(5);
    HT1632C_WR_HGH;
//    _delay_c(5);
  } while (nbits >>= 1);
}

/**
 * @brief Write Data to display buffer, no transfer to HT1632.
 * @param data pointer to data array
 * @param addr begin address
 * @param len bytes to write
 */
void display_WriteData(const uint8_t * data, uint8_t addr, uint8_t len) {
  /* check given values */
  if (len == 0) {
    return;
  }
  if (addr >= DISPLAY_COLUMNS) {
    addr = DISPLAY_COLUMNS - 1;
  }
  if ((addr + len) > DISPLAY_COLUMNS) {
    len = DISPLAY_COLUMNS - addr;
  }

#ifdef FRAMEBUFFER_ROTATE
  /* Copy given data to framebuffer */
  int i, a;
  for (i=0; i<len; i++) {
    a = addr + i;
    display_Buffer[a] = data[i];
  }
  //display_WriteBuffer();

#else
  // without dma - pool mode
  int i;

  GPIO_SPI_SW();
  HT1632C_CS_ON;
  i = (0x280 | addr);
  // write addr
  _display_WriteBits(i, 0x200);

  GPIO_SPI_HW();
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }
  for (i=0; i<len; i++) {
    while ((SPI1->SR & SPI_SR_TXE) == 0) { __NOP(); }
    SPI1->DR = data[i];
  }
  // wait for end of transmission
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }
  HT1632C_CS_OFF;

  // with dma
  uint16_t a = 0x280 | addr;

  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = data;
  /* Set DMA data transfer length (SPI buffer length). */
  DMA1_Channel3->CNDTR = len;

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits(a, 0x200);

  GPIO_SPI_HW();

  // start transfer
  SPI1->CR2 |= (SPI_CR2_TXDMAEN);
  DMA1_Channel3->CCR |= DMA_CCR_EN;
  // End of transaction in DMA1_Channel2_3_IRQHandler()
#endif /* FRAMEBUFFER_ROTATE */
}

/**
 * @brief Write all buffer to HT1632C
 */
void display_WriteBuffer(void) {
  display_WriteBuf(display_Buffer);
}

void display_WriteBuf(const uint8_t * buf) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

#ifdef FRAMEBUFFER_ROTATE
  uint8_t spi_buf[DISPLAY_COLUMNS] = {0};

  /* Rotate display buffer */
  uint8_t i, j, p = 0x01;
  for (i=0; i<8; i++) {
    if ((buf[0] & p) != 0) { spi_buf[i] |= 0x01; }
    if ((buf[1] & p) != 0) { spi_buf[i] |= 0x02; }
    if ((buf[2] & p) != 0) { spi_buf[i] |= 0x04; }
    if ((buf[3] & p) != 0) { spi_buf[i] |= 0x08; }
    if ((buf[4] & p) != 0) { spi_buf[i] |= 0x10; }
    if ((buf[5] & p) != 0) { spi_buf[i] |= 0x20; }
    if ((buf[6] & p) != 0) { spi_buf[i] |= 0x40; }
    if ((buf[7] & p) != 0) { spi_buf[i] |= 0x80; }

    j = i + 8;
    if ((buf[8] & p) != 0)  { spi_buf[j] |= 0x01; }
    if ((buf[9] & p) != 0)  { spi_buf[j] |= 0x02; }
    if ((buf[10] & p) != 0) { spi_buf[j] |= 0x04; }
    if ((buf[11] & p) != 0) { spi_buf[j] |= 0x08; }
    if ((buf[12] & p) != 0) { spi_buf[j] |= 0x10; }
    if ((buf[13] & p) != 0) { spi_buf[j] |= 0x20; }
    if ((buf[14] & p) != 0) { spi_buf[j] |= 0x40; }
    if ((buf[15] & p) != 0) { spi_buf[j] |= 0x80; }

    j = i + 16;
    if ((buf[16] & p) != 0) { spi_buf[j] |= 0x01; }
    if ((buf[17] & p) != 0) { spi_buf[j] |= 0x02; }
    if ((buf[18] & p) != 0) { spi_buf[j] |= 0x04; }
    if ((buf[19] & p) != 0) { spi_buf[j] |= 0x08; }
    if ((buf[20] & p) != 0) { spi_buf[j] |= 0x10; }
    if ((buf[21] & p) != 0) { spi_buf[j] |= 0x20; }
    if ((buf[22] & p) != 0) { spi_buf[j] |= 0x40; }
    if ((buf[23] & p) != 0) { spi_buf[j] |= 0x80; }

    p <<= 1;
  }

  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = (uint32_t)&spi_buf[0];
#else
  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = (uint32_t)&buf[0];
#endif /* FRAMEBUFFER_ROTATE */

  /* Set DMA data transfer length (SPI buffer length). */
  DMA1_Channel3->CNDTR = DISPLAY_COLUMNS;

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits(0x280, 0x200);

  GPIO_SPI_HW();

  /* start transfer */
  SPI1->CR2 |= SPI_CR2_TXDMAEN;
  DMA1_Channel3->CCR |= DMA_CCR_EN;
  /* End of transaction in DMA1_Channel2_3_IRQHandler() */
}

/**
 * @brief Fill display.
 * @param pattern for filling
 */
void display_Fill(uint8_t pattern) {
  int i;
  for (i=0; i<DISPLAY_COLUMNS; i++) {
    display_Buffer[i] = pattern;
  }
  display_WriteBuffer();
}

/**
 * @brief Wait while displaye is busy for running string
 */
void display_Wait4Display(void) {
  /* wait if display is busy */
  while (Flag.DISP_BSY != 0) { 
    RTOS_DispatchTask();
    __WFI();
   }
}

/**
  * @brief This function handles TIM16 global interrupt.
  */
void TIM16_IRQHandler(void) {
  if (TIM16->SR & TIM_SR_UIF) {
    TIM16->SR = 0; // reset flag

    display_WriteBuf(&String.Buf[String.k]);

    String.k ++;
    if (String.k > String.j) {
      String.k = 0;

      String.Repeat --;
      if (String.Repeat == 0) {
        Flag.DISP_BSY = 0;
        TIM16->CR1 &= ~(TIM_CR1_CEN);
      }

    }
  }
}
