/**
 *  Author(s)...: Pashgan    http://ChipEnable.Ru
 *  Target(s)...: any MCU
 *  Compiler....: IAR 5.11A
 *  Description.: Template for tables based event system
 *  Data........: 30.09.12
*/

#include "event-system.h"

static void EmptyFunc(void);

/* ring buffer */
static volatile es_event_t cycleBuf[SIZE_BUF];
static volatile int tailBuf = 0;
static volatile int headBuf = 0;
static volatile int countBuf = 0;

static volatile es_state_t _State;

typedef struct {
  es_state_t state;
  es_event_t event;
  es_state_t nextState;
  void (*pStateFunc1)(void);
  void (*pStateFunc2)(void);
} table_state_t;

/**
  * Sate Table - heart of state machine
  */
const table_state_t stateTable[] = {
  /* STATE    EVENT   NEXT STATE  STATE_FUNC1 STATE_FUNC2 */
  /* show date, time, ... */
  // 'Hour+' Button pressed
  {stShowTime, evBTN1Pressed, stShowDate, show_Date, NULL},
  //{stShowMMSS, evBTN1Pressed, stShowDate, show_Date, NULL},
  // 'Min+' Button pressed
  {stShowTime, evSwitch2On,  stShowMMSS, show_MMSS, NULL},
  {stShowMMSS, evSwitch2Off, stShowTime, show_Time, NULL},

  /* display wdt events */
  {stShowDate,  evDisplayWDT, stShowTime, show_Time, NULL},

  /* refresh output */
  {stShowTime, evNewSecond,  stNoChange, show_Time,   NULL},
  {stShowMMSS, evNewSecond,  stNoChange, show_MMSS,   NULL},

  /* set time */
  {stShowTime, evBTN1Holded,  stSetHH,    setTimeBegin, setHHBegin},
  {stSetHH,    evBTN1Holded,  stShowTime, setTimeEnd,   show_Time},
  {stSetMM,    evBTN1Holded,  stShowTime, setTimeEnd,   show_Time},

  {stSetHH,    evBTN1Pressed, stSetMM,    setMMBegin,   NULL},
  {stSetMM,    evBTN1Pressed, stSetHH,    setHHBegin,   NULL},

  {stSetHH,    evBTN2Pressed, stNoChange, setHHInc,     setTimeShow},
  {stSetHH,    evBTN3Pressed, stNoChange, setHHDec,     setTimeShow},
  {stSetMM,    evBTN2Pressed, stNoChange, setMMInc,     setTimeShow},
  {stSetMM,    evBTN3Pressed, stNoChange, setMMDec,     setTimeShow},

  {stSetHH,    evDisplayWDT,  stShowTime, show_Time,     NULL},
  {stSetMM,    evDisplayWDT,  stShowTime, show_Time,     NULL},

  /* set calendar */
  // begin
  {stShowDate, evBTN1Holded, stSetWD,   setDateBegin, setWDBegin},
  // end
  {stSetWD,   evBTN1Holded, stShowDate, setDateEnd, show_Date},
  // next
  {stSetWD,   evBTN1Pressed, stSetMD,   setMDBegin,    NULL},
  {stSetMD,   evBTN1Pressed, stSetMon,  setMonthBegin, NULL},
  {stSetMon,  evBTN1Pressed, stSetYear, setYearBegin,  NULL},
  {stSetYear, evBTN1Pressed, stSetWD,   setWDBegin,    NULL},
  // change
  {stSetWD,   evBTN2Pressed, stNoChange, setIncWDay,  setWDShow},
  {stSetWD,   evBTN3Pressed, stNoChange, setDecWDay,  setWDShow},
  {stSetMD,   evBTN2Pressed, stNoChange, setIncMDay,  setDMShow},
  {stSetMD,   evBTN3Pressed, stNoChange, setDecMDay,  setDMShow},
  {stSetMon,  evBTN2Pressed, stNoChange, setIncMonth, setDMShow},
  {stSetMon,  evBTN3Pressed, stNoChange, setDecMonth, setDMShow},
  {stSetYear, evBTN2Pressed, stNoChange, setIncYear,  setYearShow},
  {stSetYear, evBTN3Pressed, stNoChange, setDecYear,  setYearShow},
  // WDT
  {stSetWD,   evDisplayWDT, stShowTime, show_Time, NULL},
  {stSetMD,   evDisplayWDT, stShowTime, show_Time, NULL},
  {stSetMon,  evDisplayWDT, stShowTime, show_Time, NULL},
  {stSetYear, evDisplayWDT, stShowTime, show_Time, NULL},

  /* mandatory empty table row */
  {stNoChange, eventNull, stNoChange, EmptyFunc, EmptyFunc}
};

/**
  * @brief  Take event.
  * @param  None
  * @retval Event code
  */
es_event_t ES_GetEvent(void)
{
  es_event_t event;

  if (countBuf > 0) {
    event = cycleBuf[headBuf];
    countBuf --;
    headBuf = (headBuf + 1) & (SIZE_BUF - 1);
    return event;
  }

  return eventNull;
}

/**
  * @brief  Place event.
  * @param  Event
  * @retval None
  */
void ES_PlaceEvent(es_event_t event)
{
  if (countBuf < SIZE_BUF) {
    cycleBuf[tailBuf] = event;
    tailBuf = (tailBuf + 1) & (SIZE_BUF - 1);
    countBuf ++;
  } else {
    /* event buffer overflow alarm */
    display_String("ES Error: No space for Events!\0", 3);
    while(1);
  }
}

/**
  * @brief  Initialize event system.
  * @param  Start Event
  * @retval None
  */
void ES_Init(es_state_t init_state)
{
  tailBuf = 0;
  headBuf = 0;
  countBuf = 0;
  _State = init_state;
}

/**
  * @brief  Fake function.
  * @param  None
  * @retval None
  */
static void EmptyFunc(void)
{
}

/**
  * @brief  Dispatcher of event system.
  * @param  Event
  * @retval None
  */
void ES_Dispatch(es_event_t event)
{
  int i;
  void (*pStateFunc1)(void);
  void (*pStateFunc2)(void);

  pStateFunc1 = NULL;
  pStateFunc2 = NULL;

  /* get next state */
  for (i=0; stateTable[i].state; i++) {
    /* if the current state and the received event match */
    if (stateTable[i].state == _State && stateTable[i].event == event) {
      /* change current state if needed */
      if (stateTable[i].nextState != stNoChange) {
        _State = stateTable[i].nextState;
      }
      /* get functions */
      pStateFunc1 = stateTable[i].pStateFunc1;
      pStateFunc2 = stateTable[i].pStateFunc2;
      break;
    }
  }

  if (pStateFunc1) {
    pStateFunc1();
  }

  if (pStateFunc2) {
    pStateFunc2();
  }
}

/**
  * @brief  Return current state code.
  * @param  None
  * @retval State code
  */
es_state_t ES_GetState(void)
{
  return _State;
}

/**
  * @brief  Set current state to given code .
  * @param  State
  * @retval None
  */
void ES_SetState(es_state_t new_state)
{
  _State = new_state;
}
