/**
  ******************************************************************************
  * @file           : i2c.c
  * @brief          : I2C base functions
  ******************************************************************************
  */

#include "main.h"

/* Private macros */
#define ENABLE_TIM  TIM17->CNT=1; TIM17->CR1 |= TIM_CR1_CEN
#define TIM_CNT     TIM17->CNT

/* Private variables */
static i2c_status_t st_Read = I2C_Ret_OK;

/* Private function prototypes */
static i2c_status_t i2c_check_err(void);

/**
 * @brief Check I2C bus for errors.
 * @retval I2C return code
 */
static i2c_status_t i2c_check_err(void) {
  i2c_status_t r = I2C_Ret_OK;

  if ((I2C1->ISR & I2C_ISR_NACKF) != 0) {
  /* device not present */
    r = I2C_Ret_NACK;
  } else if ((I2C1->ISR & I2C_ISR_TIMEOUT) != 0) {
  /* timeout error */
    r = I2C_Ret_Tout;
  } else if ((I2C1->ISR & (I2C_ISR_ARLO | I2C_ISR_BERR)) != 0) {
  /* other error */
    r = I2C_Ret_Err;
  }

  if (r != I2C_Ret_OK) {
  /* restart I2C and clear flags */
    I2C1->CR1 &= ~I2C_CR1_PE;
    while ((I2C1->CR1 & I2C_CR1_PE) != 0);
    I2C1->CR1 |= I2C_CR1_PE;
  }

  return r;
}

/**
 * @brief Read len bytes from I2C bus to data by reg_addr.
 * @retval I2C return code
 */
i2c_status_t user_i2c_read(const uint8_t id, const uint8_t reg_addr, uint8_t *data, uint16_t len) {
  /* wait for i2c */
  while ( I2C1->ISR & I2C_ISR_BUSY ) { __NOP(); };

  st_Read = I2C_Ret_OK;

  Flag.I2C_RX_End = 0;
  Flag.I2C_RX_Err = 0;
  Flag.I2C_TX_Err = 0;

  /* prepare i2c for sending reg addr */
  I2C1->CR2 &= ~(I2C_CR2_SADD | I2C_CR2_NBYTES | I2C_CR2_RD_WRN);
  I2C1->CR2 |= (id | (1 << I2C_CR2_NBYTES_Pos));
  I2C1->CR2 |=  (I2C_CR2_START);
  while ((I2C1->CR2 & I2C_CR2_START)) {}

  /* wait for byte request or any error */
  while ((I2C1->ISR & (I2C_ISR_TIMEOUT | I2C_ISR_ARLO | I2C_ISR_BERR | I2C_ISR_NACKF | I2C_ISR_TXE)) == 0) {
    __NOP();
  }

  if ((I2C1->ISR & I2C_ISR_TXE) != 0) {
  /* device ok, send reg addr */
    I2C1->TXDR = reg_addr;
  } else {
    st_Read = i2c_check_err();
    Flag.I2C_TX_Err = 1;
    Flag.I2C_RX_End = 1;
    return st_Read;
  }

  /* wait for i2c or any error */
  while (((I2C1->ISR & I2C_ISR_BUSY) != 0) && ((I2C1->ISR & (I2C_ISR_ARLO | I2C_ISR_BERR | I2C_ISR_NACKF)) == 0)) { __NOP(); };
  st_Read = i2c_check_err();
  if (st_Read != I2C_Ret_OK) {
    Flag.I2C_TX_Err = 1;
    Flag.I2C_RX_End = 1;
    return st_Read;
  }

  /* prepare i2c for receiving data */
  I2C1->CR2 &= ~(I2C_CR2_SADD | I2C_CR2_NBYTES | I2C_CR2_RD_WRN);
  I2C1->CR2 |= (id | (len << I2C_CR2_NBYTES_Pos) | I2C_CR2_RD_WRN);
  /* launch receiving */
  I2C1->CR2 |=  (I2C_CR2_START);
  while ((I2C1->CR2 & I2C_CR2_START)) {}

  /* receiving data */
  while (len) {
    while ((I2C1->ISR & (I2C_ISR_NACKF | I2C_ISR_RXNE))==0) { __NOP(); }
    if (I2C1->ISR & I2C_ISR_RXNE) { // Reading data
      *data = I2C1->RXDR;
      data ++;
      len --;
    } else {
      st_Read = i2c_check_err();
      Flag.I2C_RX_Err = 1;
      Flag.I2C_RX_End = 1;
      return st_Read;
    }
  }

  Flag.I2C_RX_End = 1;
  return st_Read;
}

/**
 * @brief Write len bytes to I2C bus from data by reg_addr.
 * @retval I2C return code
 */
i2c_status_t user_i2c_write(const uint8_t id, const uint8_t reg_addr, uint8_t *data, uint16_t len) {
  i2c_status_t r = I2C_Ret_OK;

  while ( I2C1->ISR & I2C_ISR_BUSY );

  Flag.I2C_TX_End = 0;
  Flag.I2C_TX_Err = 0;

  I2C1->CR2 &= ~(I2C_CR2_SADD | I2C_CR2_NBYTES | I2C_CR2_RD_WRN);
  I2C1->CR2 |= (id | ((len + 1) << I2C_CR2_NBYTES_Pos ));
  I2C1->CR2 |=  (I2C_CR2_START);
  while ((I2C1->CR2 & I2C_CR2_START)) {}

  while ((I2C1->ISR & (I2C_ISR_ARLO | I2C_ISR_BERR | I2C_ISR_NACKF | I2C_ISR_TXE)) == 0) { __NOP(); }
  if ((I2C1->ISR & I2C_ISR_TXE) != 0) {
    I2C1->TXDR = reg_addr;
  } else {
    r = i2c_check_err();
    Flag.I2C_TX_Err = 1;
    Flag.I2C_TX_End = 1;
    return r;
  }

  while (len) {
  while ((I2C1->ISR & (I2C_ISR_NACKF | I2C_ISR_TXE)) == 0) { __NOP(); }
    if ((I2C1->ISR & I2C_ISR_TXE) != 0) {
      I2C1->TXDR = *data;
      data ++;
      len --;
    } else {
      r = i2c_check_err();
      Flag.I2C_TX_Err = 1;
      Flag.I2C_TX_End = 1;
      return r;
    }
  }

  Flag.I2C_TX_End = 1;
  return r;
}
